#include "decision_tree.hpp"
#include <boost/foreach.hpp>

namespace Tree_AG{

  // Split the dataset of Rows comparing column, with value stored in node
  //
  // @param current     (current row indexes)
  // @param rows left   (OUT: left side of the split)
  // @param rows right  (OUT: right side of the split)
  // @param colum_index (column index where to split)
  // @param value       (comparison value);
  //
  void DecisionTree::divide_set(Rows_indexes& current_indexes, 
				Rows_indexes& left, Rows_indexes& right,
				unsigned int column_index, Value& value){
   
    BOOST_FOREACH(Rows_indexes::value_type& i, current_indexes){
      
      //std::cout << "Considering index=" << i 
      //		<< " column index=" << column_index 
      //		<< " rows_[i][column_index]=" 
      //		<< rows_[i][column_index] << std::endl;
      
      if (compare(rows_[i][column_index], value))
	left.push_back(i);
      else
	right.push_back(i);      
    }  
  }

  // Build tree
  // @param score_function
  //
  Node * DecisionTree::build_tree(Rows_indexes& indexes,
				  float(*score_function)(Rows&, Rows_indexes&))
  {
    if (indexes.size() == 0)
      return new Node();
    
    Dictionary d;
    Value best_value;
    Rows_indexes best_left_indexes, best_right_indexes;
    float best_gain, current_score, score_left, score_right, p, gain;    
    unsigned int best_column, column_size;

    column_size = indexes.size();
    best_gain = 0.0;
    current_score = score_function(rows_, indexes);
    
    std::cout << "Working on rows =(" << indexes << ")" << std::endl;
    std::cout << "Current score: " << current_score << std::endl;

    // FOREACH column, 
    for(unsigned int i=0; i<column_size; i++){

      d.clear(); 

      std::cout << std::endl << std::endl;
      // FOREACH row 
      BOOST_FOREACH(Rows::value_type r, rows_){
	//	std::cout << "column:"<<indexes[i]<<" row["
	//	  <<indexes[i]<<"]="<<r[indexes[i]]<<std::endl;
	d.add(r[indexes[i]]);  // load a dictionary
      }// each row
    
      // FOREACH word in dictionary
      Dictionary::const_iterator it = d.begin();
      Dictionary::const_iterator it_end = d.end();
      for (; it!=it_end; ++it){
	Rows_indexes left_indexes, right_indexes;
	Value key = it->first;

	// divide set on colum_index, key and compute left, right score
	//
	divide_set(indexes, left_indexes, right_indexes, indexes[i], key);
	score_left=score_function(rows_, left_indexes);
	score_right=score_function(rows_, right_indexes);

	// compute information gain
	//
	p = ((float)left_indexes.size()/rows_.size());
	gain = current_score - ( p * score_left ) - ( (1 - p) * score_right);
	
	//
	std::cout << "For (column=" << i << ", key=" << key << ")" 
		  << std::endl << "left: ";
	BOOST_FOREACH(Rows_indexes::value_type index, left_indexes){
	  std::cout << index << "=(" << rows_[index] << ") ";
	}
	std::cout << std::endl << "right: ";
	BOOST_FOREACH(Rows_indexes::value_type index, right_indexes){
	  std::cout << index << "=(" <<rows_[index] << ") ";
	}
	std::cout << std::endl << " score left:" << score_left 
		  << " score right:" << score_right 
		  << " p:" << p << " gain:" << gain << std::endl;
	//

	if (gain>best_gain && 
	    left_indexes.size()>0 && right_indexes.size()>0){
	  
	  std::cout << " Improving gain=" << gain 
		    << " best_gain=" << best_gain
		    << " best_left_indexes=" << left_indexes 
		    << " best_right_indexes=" << right_indexes 
		    << " best_value=" << key 
		    << " best_column=" << indexes[i]
		    << std::endl << std::endl;
	  
	  best_gain = gain;                    // store best gain
	  best_left_indexes = left_indexes;    // store best left_indexes
	  best_right_indexes = right_indexes;  // store best right_indexes
	  best_value = key;                    // store best_key
	  best_column = indexes[i];          // store best_column 	    
	}
      } // for each word in dictionary
    } // for each column

    if (best_gain>0){                        // another split and inner node
      
      std::cout << std::endl << std::endl << "//// Recursing " 
		<< " best_gain=" << best_gain << std::endl
		<< " best_left_indexes=" << best_left_indexes
		<< " best_right_indexes=" << best_right_indexes
		<< " best_value=" << best_value
		<< " best_column=" << best_column<< std::endl;
      
      std::cout << "/// LEFT" << std::endl;
      Node * left_child = build_tree(best_left_indexes);

      std::cout << "/// RIGHT" << std::endl;
      Node * right_child = build_tree(best_right_indexes);

      std::cout << "end recursion, creating node" << std::endl;
      return new Node(best_column, best_value,
		      left_child, right_child);
    
    } else {          // leaf

      Dictionary d;
       // FOREACH column, 	// FOREACH row 
      for(unsigned int i=0; i<column_size; i++){
	BOOST_FOREACH(Rows::value_type r, rows_){
	  d.add(r[indexes[i]]);  // load a dictionary	
 	}// each row
      }
      return new Node(d);
    }
  } // build tree

  
  Dictionary classify(Row& new_observation, Node * node)
  {
    Dictionary d;
    Value v_obs, v_node;
    int c;

    if (node->is_leaf())
      {
	d= node->get_dictionary();
	std::cout << "Classified Dictionary " << d << std::endl;
	return d;
      }
    else
      {
	c = node->get_column();    // get the split column of this node
	v_obs = new_observation[c];// and extract that future from obs  
	v_node = node->get_value();// and from the node

	std::cout << "Node getting column=" << c 
		  << " in new_observation the value is=" << v_obs
		  << " in tree the value is=" << v_node << std::endl;

	if (compare(v_obs, v_node )){   // TRUE BRANCH
	  
	  std::cout << " True Branch" << std::endl;
	  return classify(new_observation, &(node->right_child()));

	} else {  // FALSE BRANCH

	  std::cout << " False Branch" << std::endl;
	  return classify(new_observation, &(node->left_child()));

	}
      }
    }

};
