#ifndef TREE_H
#define TREE_H 1

#include <boost/ptr_container/ptr_array.hpp>
#include <iostream>
#include "defs.hpp"
#include "criteria.hpp"
#include "dictionary.hpp"

namespace Tree_AG{

  //
  // A Node: a common ancestor
  //
  class Node : boost::noncopyable{

  public:
    // there are at most 2 nodes children
    typedef boost::ptr_array< boost::nullable<Node>, 2> Nodes;  
    typedef Nodes::iterator iterator;
    typedef Nodes::const_iterator const_iterator;

    Node () {};
    Node (Dictionary& d) : 
      column_index_(-1), value_(-1), d_(d){};
    Node (Value & v) : value_(v) {};
    Node (unsigned int column_index, Value& value, Node * l, Node * r) :
      column_index_(column_index), value_(value)
    {left_child(l); right_child(r);}

    // node access
    //
    inline Node& left_child(Node * n)
    { nodes_.replace(0, n); return nodes_[0];}
    inline Node& right_child(Node * n)
    { nodes_.replace(1, n); return nodes_[1];}
    inline Node& left_child(){ return nodes_[0];}
    inline Node& right_child(){ return nodes_[1];}
    
    // node delete and esistance check
    //
    inline void delete_left_child()
    { std::auto_ptr<Node> x; nodes_.replace(0,  x);}
    inline void delete_right_child()
    { std::auto_ptr<Node> x; nodes_.replace(0,  x);}

    inline bool empty_left_child() const { return nodes_.is_null(0);}
    inline bool empty_right_child() const { return nodes_.is_null(1);}
    bool is_leaf() const { return empty_left_child() && empty_right_child();}

    // nodes traversal 
    //
    inline iterator begin(){ return nodes_.begin(); }
    inline const_iterator begin() const { return nodes_.begin();}
    inline iterator end(){ return nodes_.end(); }
    inline const_iterator end() const { return nodes_.end();}    

    // node attributes
    //
    inline Dictionary get_dictionary(void) const { return d_; }
    inline int get_column(void) const { return column_index_; }
    inline Value get_value(void) const { return value_; }

  private:  
    int column_index_; // colum index of the criteria to test
    Value value_; // the attribute where this node is splitting
    Dictionary d_; // a dictionary of arithmetic, categorical values
    Nodes nodes_; // there are two possible decisions (yes, no)    
    
    // access val stored
    friend std::ostream& operator << (std::ostream& os, const Node & n);
  };

  //
  ///// FREE METHODS  ////////////////////////////////////////////


  // Depth first node
  // @param node n
  //
  void depth_first(const Node& n, unsigned int level=0);  
};


#endif
